import ast
from typing import Tuple

Imports = list[Tuple[str, str]]


class ImportsGenerator:
    def __init__(self):
        self.imports: Imports = []

    def add(self, module, obj):
        if not self.exists(obj):
            self.imports.append((module, obj))

    def exists(self, obj: str):
        return any(i[1] == obj for i in self.imports)

    def add_bulk(self, module, objs: list[str]):
        for obj in objs:
            self.add(module, obj)

    def gen(self):
        merged_imports = self.get_merge_imports()
        result = ""
        for module, objs in merged_imports.items():
            if module is None:
                result += f"import {', '.join(objs)}\n"
            else:
                result += f"from {module} import {', '.join(objs)}\n"

        return result

    def get_merge_imports(self):
        merged_imports = {}
        for module, obj in self.imports:
            if module not in merged_imports.keys():
                merged_imports[module] = [obj]
            else:
                merged_imports[module].append(obj)
        return merged_imports

    def parse_imports(self, code):
        imports = []
        tree = ast.parse(code)

        for node in ast.walk(tree):
            if isinstance(node, ast.Import):
                for alias in node.names:
                    self.add(None, alias.name)
            elif isinstance(node, ast.ImportFrom):
                module = node.module
                for alias in node.names:
                    if module:
                        self.add(f"{'.' * node.level}{module}", alias.name)

        return imports

    @property
    def imported_classes(self):
        for _import in self.imports:
            yield _import[1]


class ViewImportsGenerator(ImportsGenerator):
    def add_login_required(self):
        self.add("django.contrib.auth.mixins", "LoginRequiredMixin")

    def add_user_test(self):
        self.add("django.contrib.auth.mixins", "UserPassesTestMixin")

    def add_reverse(self):
        self.add("django.shortcuts", "reverse")

    def add_reverse_lazy(self):
        self.add("django.shortcuts", "reverse_lazy")



from django.urls import path

from .views import AddEqualSignView, ExampleCreateView, ExampleDetailView, ExampleListView


app_name = "example"
urlpatterns = [
    path("<int:pk>/add-equal-sign/", AddEqualSignView.as_view(), name="add_equal_sign"),
    path("create/", ExampleCreateView.as_view(), name="create"),
    path("<int:pk>/", ExampleDetailView.as_view(), name="detail"),
    path("", ExampleListView.as_view(), name="list")
]    



import os

from pytest import fixture
from django.apps import apps

from hogwarts.magic_urls.base import import_views
from hogwarts.magic_urls.gen_urls import gen_path, UrlGenerator
from hogwarts.magic_urls import custom_path

from hogwarts import _test_views
from hogwarts.utils import code_strip


def test_it_generates_path():
    result = gen_path(_test_views.MyListView, "my")
    expected = 'path("", MyListView.as_view(), name="list")'

    assert result == expected


def test_it_generates_path_for_function():
    result = gen_path(_test_views.confirm_post_view, "none")
    expected = 'path("confirm-post/", confirm_post_view, name="confirm_post")'

    assert result == expected


def test_it_extracts_metadata():
    @custom_path("green", "green-hello/")
    class RedView:
        pass

    result = gen_path(RedView, "none")
    expected = 'path("green-hello/", RedView.as_view(), name="green")'

    assert result == expected


views = import_views(_test_views)


@fixture
def generator():
    base_url = apps.get_app_config("hogwarts").path
    return UrlGenerator(_test_views, os.path.join(base_url, "urls.py"), "my", True)


def test_it_generates_urls(generator):
    result = generator.gen_urlpatterns(views)

    expected = """
    urlpatterns = [
        path("form/", MyFormView.as_view(), name="form"),
        path("", MyListView.as_view(), name="list"),
        path("confirm-post/", confirm_post_view, name="confirm_post"),
        path("get/", get_view, name="get"),
        path("post/", post_view, name="post")
    ]"""

    assert result == code_strip(expected)


def test_it_generates_imports(generator):
    result = generator.gen_url_imports([
        _test_views.MyListView,
        _test_views.MyFormView,
        _test_views.get_view
    ])

    expected = """
        from django.urls import path
    
        from .views import MyListView, MyFormView, get_view
    """

    assert code_strip(result) == code_strip(expected)



from hogwarts.magic_views.gen_imports import ImportsGenerator
from hogwarts.utils import code_strip

code = """
import ast
import datetime

from django.views.generic import DetailView, ListView
from .forms import MyForm
"""


def test_it_generates_imports_from_code():
    gen = ImportsGenerator()
    expected = {
        None: ["ast", "datetime"],
        "django.views.generic": ["DetailView", "ListView"],
        ".forms": ["MyForm"],
    }
    gen.parse_imports(code_strip(code))

    assert gen.get_merge_imports() == expected


def test_it_merges_import_with_code():
    gen = ImportsGenerator()

    expected_code = """
    import ast, datetime
    from django.views.generic import DetailView, ListView, UpdateView
    from .forms import MyForm, YouForm
    """

    gen.parse_imports(code_strip(code))
    gen.add("django.views.generic", "UpdateView")
    gen.add(".forms", "YouForm")

    assert gen.gen() == code_strip(expected_code)



from typing import Optional
from rich.console import Console

from ..utils import to_plural, code_strip, remove_empty_lines, parse_class_names, remove_imports
from .gen_imports import ViewImportsGenerator


console = Console()


class ViewGenerator:
    def __init__(
            self,
            model,
            smart_mode=False,
            model_is_namespace=False,
            code: Optional[str] = None
    ):
        self.smart_mode = smart_mode
        self.model_is_namespace = model_is_namespace

        self.model_name = model.__name__
        self.name = model.__name__.lower()
        self.fields = [field.name for field in model._meta.fields if field.editable]
        if "id" in self.fields:
            self.fields.remove("id")

        self.creator_field = None
        self.code = code
        self.merge = code is not None
        self.existing_class_names = []

        if self.code:
            self.existing_class_names = parse_class_names(self.code)

        for field in self.fields:
            if field in ["user", "author", "owner", "creator"]:
                self.fields.remove(field)
                self.creator_field = field
                break

        self.imports_generator = ViewImportsGenerator()
        self.generic_views = []

    def gen(self):
        detail = self.detail()
        _list = self.list()
        create = self.create()
        update = self.update()

        result = code_strip(self.gen_imports())
        if self.merge:
            result += "\n"
            result += remove_imports(self.code)

        for view in [detail, _list, create, update]:
            if view is not None:
                result += f"\n\n{code_strip(view)}"
                console.print(f"generated {view.splitlines()[0]}", style="bright_black")
        return result

    def gen_imports(self):
        if self.merge:
            self.imports_generator.parse_imports(self.code)

        self.imports_generator.add_bulk("django.views.generic", self.generic_views)
        self.imports_generator.add(".models", self.model_name)

        return self.imports_generator.gen()

    def create(self):
        if self.merge and f"{self.model_name}CreateView" in self.existing_class_names:
            return None

        self.generic_views.append("CreateView")
        builder = self.get_builder("create")

        builder.set_fields(self.fields)
        self.set_template(builder, "create")

        if self.smart_mode:
            self.imports_generator.add_login_required()
            builder.set_class("create", ["LoginRequiredMixin"])

            if self.creator_field:
                function = f"""
                def form_valid(self, form):
                    form.instance.{self.creator_field} = self.request.user
                    return super().form_valid(form)
                """
                builder.set_extra_code(code_strip(function))

        if self.model_is_namespace:
            self.imports_generator.add_reverse()
            function = f"""
            def get_success_url(self):
                return reverse("{to_plural(self.name)}:detail", args=[self.object.id])
            """
            builder.set_extra_code(code_strip(function))
        else:
            builder.set_success_url("/")

        return builder.gen()

    def update(self):
        if self.merge and f"{self.model_name}UpdateView" in self.existing_class_names:
            return None

        self.generic_views.append("UpdateView")
        builder = self.get_builder("update")
        builder.set_fields(self.fields)
        self.set_template(builder, "update")

        if self.smart_mode:
            if self.creator_field:
                self.imports_generator.add_user_test()
                builder.set_class("update", ["UserPassesTestMixin"])
                function = """
                def test_func(self):
                    return self.get_object() == self.request.user
                """
                builder.set_extra_code(code_strip(function))
            else:
                self.imports_generator.add_login_required()
                builder.set_class("update", ["LoginRequiredMixin"])

        if self.model_is_namespace:
            self.imports_generator.add_reverse()
            function = f"""
            def get_success_url(self):
                return reverse("{to_plural(self.name)}:detail", args=[self.get_object().id])
            """
            builder.set_extra_code(code_strip(function))
        else:
            builder.set_success_url("/")

        return builder.gen()

    def detail(self):
        if self.merge and f"{self.model_name}DetailView" in self.existing_class_names:
            return None

        self.generic_views.append("DetailView")
        builder = self.get_builder("detail")
        builder.set_context_object_name(self.name)
        self.set_template(builder, "detail")

        return builder.gen()

    def list(self):
        if self.merge and f"{self.model_name}ListView" in self.existing_class_names:
            return None

        self.generic_views.append("ListView")
        builder = self.get_builder("list")
        builder.set_context_object_name(to_plural(self.name))
        self.set_template(builder, "list")

        return builder.gen()

    def get_builder(self, action, mixins=[]):
        builder = ClassViewBuilder(self.model_name)
        builder.set_class(action, mixins)
        builder.set_model()

        return builder

    def set_template(self, builder, action):
        builder.set_template_name(f"{to_plural(self.name)}/{self.name}_{action}.html")


class ClassViewBuilder:
    def __init__(self, model_name):
        self.model_name = model_name
        self.result = {
            "class": "",
            "model": "",
            "fields": "",
            "context_object_name": "",
            "template_name": "",
            "success_url": "",
        }

        self.extra_codes = []

    def set_class(self, action, mixins=[]):
        action = action.capitalize()
        class_name = f"{self.model_name}{action}View"
        inherits = [*mixins, f"{action}View"]

        self.result["class"] = f"class {class_name}({', '.join(inherits)}):\n"

    def set_model(self):
        self.result["model"] = f'    model = {self.model_name}\n'

    def set_fields(self, fields: list[str]):
        fields = [f'"{field}"' for field in fields]
        self.result["fields"] = f"    fields = [{', '.join(fields)}]\n"

    def set_context_object_name(self, name):
        self.result["context_object_name"] = f'    context_object_name = "{name}"\n'

    def set_template_name(self, name):
        self.result["template_name"] = f'    template_name = "{name}"\n'

    def set_success_url(self, url):
        self.result["success_url"] = f'    success_url = "{url}"\n'

    def set_extra_code(self, extra_code):
        extra_code = "\n".join(map(lambda line: " " * 4 + line, extra_code.splitlines()))
        extra_code = remove_empty_lines(extra_code)

        self.extra_codes.append(f"\n{extra_code}\n")

    def gen(self):
        code = ""

        for key in self.result.keys():
            code += self.result[key]

        for extra in self.extra_codes:
            code += extra

        return code



from django.urls import path
from django.views import View

from hogwarts import _test_views
from hogwarts._test_views import MyFormView, MyListView, get_view, post_view, confirm_post_view
from hogwarts.magic_urls.auto_urls import (
    get_path,
    get_path_name,
    import_views,
)
from hogwarts.magic_urls import custom_path
from hogwarts.magic_urls.base import get_path_url


class ProductCreateView(View): pass
class ProductsListView(View): pass
class ProductDetailView(View): pass
class SendMessageView(View): pass
class ConfirmOrderView(View): pass
def product_update_view(): pass
def dumb_files_view(): pass


def test_it_imports_user_defined_views():
    imported_classes = import_views(_test_views)
    expected = [MyFormView, MyListView, confirm_post_view, get_view, post_view]

    assert imported_classes == expected


def test_crud_get_path_name():
    payload = [
        (ProductCreateView, 'create'),
        (ProductsListView, 'list'),
        (ProductDetailView, 'detail'),
        (product_update_view, 'update')
    ]

    for view, expected_path_name in payload:
        path_name = get_path_name(view, 'products')
        assert path_name == expected_path_name


def test_detail_path_url():
    path_name = get_path_name(ProductDetailView, 'products')
    path_url = get_path_url(path_name, 'products')

    assert path_url == "<int:pk>/"


def test_get_path_name():
    payload = [
        (SendMessageView, "send_message"),
        (ConfirmOrderView, "confirm_order"),
        (dumb_files_view, "dumb_files")
    ]

    for view, expected_path_name in payload:
        path_name = get_path_name(view)
        assert path_name == expected_path_name


def test_auto_path_decorator():
    @custom_path("confirm", detail=True)
    class Some(View):
        pass

    expected_path = path("<int:pk>/confirm/", Some.as_view(), name="confirm")
    current_path = get_path(Some)

    assertPathEqual(expected_path, current_path)


def test_decorator_custom_url():
    path_url = "<int:pk>/products"

    @custom_path("confirm", path_url=path_url, detail=True)
    class Another(View):
        pass

    expected_path = path(path_url, Another.as_view(), name="confirm")
    current_path = get_path(Another)

    assertPathEqual(expected_path, current_path)


def assertPathEqual(path1, path2):
    assert path1.name == path2.name
    assert path1.pattern.regex == path2.pattern.regex



import os
import re
from inspect import isclass
from typing import Tuple, Optional

from hogwarts.magic_urls.base import import_views, get_path_name, get_path_url, view_is_detail
from .decorators import PathDecorator


class BaseUrl:
    def __init__(self, views_module, urls_path: str, app_name, force_app_name=False, single_import=False):
        self.views_module = views_module
        self.urls_path = urls_path
        self.app_name = None
        self.single_import = single_import

        if force_app_name:
            self.app_name = app_name
        else:
            self.app_name = read_app_name(self.urls_path) or app_name

    def write(self, imports, urlpatterns):
        code = f'{imports}\n\napp_name = "{self.app_name}"\n{urlpatterns}\n'

        with open(f"{self.urls_path}", 'w') as file:
            file.write(code)


class UrlGenerator(BaseUrl):
    def gen_urls_py(self):
        """
        generates code for urls.py
        fully overrides any existing code
        """
        views = import_views(self.views_module)

        imports = self.gen_url_imports(views)
        urlpatterns = self.gen_urlpatterns(views)

        self.write(imports, urlpatterns)

    def gen_url_imports(self, views: list[object]):
        if self.single_import:
            return "from django.urls import path\n\nfrom . import views"
        view_names = ", ".join(view.__name__ for view in views)
        return f"from django.urls import path\n\nfrom .views import {view_names}\n"

    def gen_urlpatterns(self, views):
        paths = []

        for view in views:
            paths.append(gen_path(view, self.app_name, self.single_import))

        paths_string = ",\n    ".join(paths)
        result = f'urlpatterns = [\n    {paths_string.strip()}\n]'

        return result


class UrlMerger(BaseUrl):
    def merge_urls_py(self):
        """
        adds new paths without touching existing paths
        """
        file = open(self.urls_path, "r")
        code = file.read()

        imports, urlpatterns = separate_imports_and_urlpatterns(code)
        views = import_views(self.views_module)

        imports = self.merge_url_imports(imports, urlpatterns, views)
        urlpatterns = self.merge_urlpatterns(urlpatterns, views)

        self.write(imports, urlpatterns)

    def merge_url_imports(self, imports, urlpatterns, views):
        if not self.single_import:
            for view in views:
                if view.__name__ not in urlpatterns:
                    imports = append_view_into_imports(imports, view.__name__)

        return imports

    def merge_urlpatterns(self, urlpatterns, views):
        paths: list[Tuple[str, str]] = []

        for view in views:
            path = gen_path(view, self.app_name, self.single_import)
            paths.append((path, view.__name__))

        for path in paths:
            if path[1] not in urlpatterns:
                urlpatterns = append_path_into_urlpatterns(urlpatterns, path[0])

        return urlpatterns


def gen_path(view, app_name, from_view_file=False) -> str:
    decorator = PathDecorator(view)
    if decorator.exists():
        path_name = decorator.get_path_name()
        path_url = decorator.get_path_url()

        print(
            f"info: You have provided @custom_path decorator in {view.__name__}"
            f"don't forget to remove because it is not in use anymore"
        )
    else:
        path_name = get_path_name(view, app_name)
        model = getattr(view, "model", False)
        path_url = get_path_url(
            path_name,
            model_name=model.__name__ if model else "none",
            detail=view_is_detail(view) if isclass(view) else False
        )

    view_name = view.__name__
    view_function = f"{view_name}.as_view()" if isclass(view) else view_name
    view_function = f"views.{view_function}" if from_view_file else view_function

    return f'path("{path_url}", {view_function}, name="{path_name}")'


# Warning. below are purely algorithmic functions, no need to read


def separate_imports_and_urlpatterns(code: str) -> Tuple[str, str]:
    imports = ""
    urlpatterns = ""

    lines = code.splitlines()

    for i in range(len(lines)):
        if lines[i].startswith("from"):
            imports += lines[i] + "\n"

            if lines[i].endswith("("):
                x = i + 1
                while not lines[x].endswith(")"):
                    imports += lines[x] + "\n"
                    x += 1
                imports += lines[x] + "\n"

        if lines[i].startswith("urlpatterns"):
            for x in range(i, len(lines)):
                urlpatterns += lines[x] + "\n"
            break

    return imports, urlpatterns


def append_view_into_imports(imports: str, view: str):
    lines = imports.splitlines()

    if "from .views" not in imports:
        lines.append("from .views import")

    for i in range(len(lines)):
        if lines[i].startswith("from .views"):
            if lines[i].endswith("("):
                x = i + 1
                while not lines[x].endswith(")"):
                    x += 1

                if lines[x].strip() == ")":
                    lines[x - 1] += ","
                    lines[x] = lines[x].replace(")", f"    {view}\n)")
                else:
                    lines[x] = lines[x].replace(")", f", {view})")

                break

            if lines[i].strip().endswith("import"):
                lines[i] = lines[i].strip()
                lines[i] += " " + view
            else:
                tokens = lines[i].split(" ")
                if view not in tokens:
                    lines[i] += ", " + view

            break

    return "\n".join(lines)


def append_path_into_urlpatterns(urlpatterns: str, path: str):
    lines = urlpatterns.splitlines()
    lines = [line.rstrip() for line in lines]
    bracket = None

    for i, line in enumerate(lines):
        if "]" in line:
            bracket = i
            break

    if bracket:
        if not lines[bracket - 1].endswith(","):
            lines[bracket - 1] = lines[bracket - 1] + ","

        lines[bracket] = "    " + path
        lines.append("]")

    return "\n".join(lines)


def get_app_name(code: str) -> Optional[str]:
    lines = code.splitlines()

    for line in lines:
        if line.startswith("app_name"):
            pattern = r'"([^"]*)"'
            match = re.search(pattern, line)
            return match.group(1)

    return None


def urlpatterns_is_empty(code):
    if code.strip() == "" or "urlpatterns" not in code:
        return True

    result = re.search(r'urlpatterns\s*=\s*\[\s*\]', code)
    return result is not None


def read_app_name(urls_path: str):
    if not os.path.exists(urls_path):
        return None

    file = open(urls_path, "r")
    code = file.read()

    return get_app_name(code)



from hogwarts.magic_views import ViewGenerator

from hogwarts.models import Article
from hogwarts.utils import code_strip


generator = ViewGenerator(Article)

def test_it_generates_detail_view():
    code = generator.detail()
    expected_code = """
    class ArticleDetailView(DetailView):
        model = Article
        context_object_name = "article"
        template_name = "articles/article_detail.html"
    """

    assert code_strip(code) == code_strip(expected_code)


def test_it_generates_list_view():
    code = generator.list()
    expected_code = """
    class ArticleListView(ListView):
        model = Article
        context_object_name = "articles"
        template_name = "articles/article_list.html"
    """

    assert code_strip(code) == code_strip(expected_code)


def test_it_generated_create_view():
    code = generator.create()
    expected_code = """
    class ArticleCreateView(CreateView):
        model = Article
        fields = ["id", "title", "description", "beta"]
        template_name = "articles/article_create.html"
        success_url = "/"
    """

    assert code_strip(code) == code_strip(expected_code)


def test_it_generated_update_view():
    code = generator.update()
    expected_code = """
    class ArticleUpdateView(UpdateView):
        model = Article
        fields = ["id", "title", "description", "beta"]
        template_name = "articles/article_update.html"
        success_url = "/"
    """

    assert code_strip(code) == code_strip(expected_code)


smart_generator = ViewGenerator(Article, smart_mode=True)


def test_smart_mode_create_view():
    code = smart_generator.create()
    expected_code = """
    class ArticleCreateView(LoginRequiredMixin, CreateView):
        model = Article
        fields = ["id", "title", "description", "beta"]
        template_name = "articles/article_create.html"
        success_url = "/"
        
        def form_valid(self, form):
            form.instance.author = self.request.user
            return super().form_valid(form)
    """

    assert code_strip(code) == code_strip(expected_code)


def test_smart_mode_update_view():
    code = smart_generator.update()
    expected_code = """
    class ArticleUpdateView(UserPassesTestMixin, UpdateView):
        model = Article
        fields = ["id", "title", "description", "beta"]
        template_name = "articles/article_update.html"
        success_url = "/"
        
        def test_func(self):
            return self.get_object() == self.request.user
    """

    assert code_strip(code) == code_strip(expected_code)


namespace_generator = ViewGenerator(Article, model_is_namespace=True)


def test_namespace_create_view():
    code = namespace_generator.create()
    expected_code = """
    class ArticleCreateView(CreateView):
        model = Article
        fields = ["id", "title", "description", "beta"]
        template_name = "articles/article_create.html"
    
        def get_success_url(self):
            return reverse("articles:detail", args=[self.object.id])
    """

    assert code_strip(code) == code_strip(expected_code)


def test_namespace_update_view():
    code = namespace_generator.update()
    expected_code = """
    class ArticleUpdateView(UpdateView):
        model = Article
        fields = ["id", "title", "description", "beta"]
        template_name = "articles/article_update.html"
        
        def get_success_url(self):
            return reverse("articles:detail", args=[self.get_object().id])
    """

    assert code_strip(code) == code_strip(expected_code)


def test_smart_mode_and_namespace_does_not_conflict():
    universal_generator = ViewGenerator(Article, smart_mode=True, model_is_namespace=True)
    code = universal_generator.create()

    expected_code = """
    class ArticleCreateView(LoginRequiredMixin, CreateView):
        model = Article
        fields = ["id", "title", "description", "beta"]
        template_name = "articles/article_create.html"
        
        def form_valid(self, form):
            form.instance.author = self.request.user
            return super().form_valid(form)

        def get_success_url(self):
            return reverse("articles:detail", args=[self.object.id])
    """

    assert code_strip(code) == code_strip(expected_code)
    assert "LoginRequiredMixin" in universal_generator.imports_generator.gen()




import os
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Type, Union

from rich.console import Console
from jinja2 import Environment
from django.views.generic import DetailView, CreateView
from django.conf import settings
from django.db.models import Model
from django.apps import apps

from hogwarts.magic_urls.base import import_views
from hogwarts.magic_urls.utils import Path, extract_paths
from hogwarts.management.commands.base import get_views_module

GENERIC_VIEWS = ("CreateView", "UpdateView", "ListView", "DetailView")

CUSTOM_SCAFFOLD_FOLDER = getattr(settings, "HOGWARTS_SCAFFOLD_FOLDER", False)
SCAFFOLD_FOLDER = CUSTOM_SCAFFOLD_FOLDER or os.path.join(apps.get_app_config("hogwarts").path, "scaffold")

TEMPLATES_FOLDER = settings.TEMPLATES[0]["DIRS"][0]
env = Environment("[#", "#]", "[[", "]]")

console = Console()


class ViewType(Enum):
    CREATE = "CreateView"
    UPDATE = "UpdateView"
    DETAIL = "DetailView"
    LIST = "ListView"


@dataclass
class Endpoint:
    view: Union[DetailView, CreateView]
    template_name: Optional[str]
    path_name: str
    model: Optional[Type[Model]]
    view_type: Optional[ViewType]


def gen_templates(app_name: str):
    endpoints = get_endpoints(app_name)
    console.print(f"generating in templates folder [bold]{TEMPLATES_FOLDER}")
    if CUSTOM_SCAFFOLD_FOLDER:
        console.print(f"using custom scaffold {CUSTOM_SCAFFOLD_FOLDER}", style="yellow")
    else:
        console.print(f"using default scaffold folder", style="bright_black")

    for endpoint in endpoints:
        if not endpoint.model or template_exists(endpoint.template_name):
            continue

        fields = [field for field in endpoint.model._meta.fields]
        model_name = endpoint.model.__name__

        if endpoint.view_type == ViewType.CREATE:
            result = render_template({"model": model_name}, "create")
            write_template(endpoint.template_name, result)
            console.print("created template:", endpoint.template_name, style="bright_black")

        elif endpoint.view_type == ViewType.UPDATE:
            result = render_template({"model": model_name}, "update")
            write_template(endpoint.template_name, result)
            console.print("created template:", endpoint.template_name, style="bright_black")

        elif endpoint.view_type == ViewType.LIST:
            name = endpoint.view.context_object_name
            create_path_name = find_path_name(endpoints, endpoint.model, ViewType.CREATE)
            detail_path_name = find_path_name(endpoints, endpoint.model, ViewType.DETAIL)

            context_data = {
                "fields": fields,
                "item": name[:-1],
                "items": name,
                "create_path_name": create_path_name,
                "detail_path_name": detail_path_name
            }

            result = render_template(context_data, "list")
            write_template(endpoint.template_name, result)
            console.print("created template:", endpoint.template_name, style="bright_black")

        elif endpoint.view_type == ViewType.DETAIL:
            name = endpoint.view.context_object_name
            update_path_name = find_path_name(endpoints, endpoint.model, ViewType.UPDATE)
            list_path_name = find_path_name(endpoints, endpoint.model, ViewType.LIST)
            context_data = {
                "fields": fields,
                "item": name,
                "list_path_name": list_path_name,
                "update_path_name": update_path_name
            }

            result = render_template(context_data, "detail")
            write_template(endpoint.template_name, result)
            console.print("created template:", endpoint.template_name, style="bright_black")


def find_path_name(endpoints: list[Endpoint], model, view_type: ViewType):
    for endpoint in endpoints:
        if endpoint.model == model and endpoint.view_type == view_type and endpoint.path_name:
            return endpoint.path_name

    return None


def render_template(context_data: dict, action):
    create = open(os.path.join(SCAFFOLD_FOLDER, f"{action}.html"), "r").read()
    template = env.from_string(create)
    return template.render(context_data)


def write_template(template_path, content):
    full_path = os.path.join(TEMPLATES_FOLDER, template_path)

    dir_path, file_name = os.path.split(full_path)
    os.makedirs(dir_path, exist_ok=True)

    with open(full_path, 'wb') as file:
        file.write(content.encode("utf-8"))


def template_exists(template_path):
    return os.path.exists(os.path.join(TEMPLATES_FOLDER, template_path))


def get_endpoints(app_name: str):
    views = get_views(app_name)
    paths = get_paths(app_name)

    endpoints: list[Endpoint] = []

    for view in views:
        endpoint = get_endpoint(view, paths, app_name)
        if endpoint.view_type:
            endpoints.append(endpoint)

    return endpoints


def get_endpoint(view, paths: list[Path], app_name: Optional[str]):
    path_name = None

    found = False
    for path in paths:
        if path.view == view.__name__:
            path_name = path.path_name
            found = True

    if found:
        path_name = f"{app_name}:{path_name}" if app_name else path_name

    view_type = get_view_type(view.__name__)
    model = None

    if hasattr(view, 'model') and view.model is not None:
        model = view.model
    else:
        try: model = view.form_class._meta.model
        except: pass

    template = getattr(view, "template_name", None)

    return Endpoint(view, template, path_name, model, view_type)


def get_views(app_name: str):
    views_module = get_views_module(app_name)
    return import_views(views_module)


def get_paths(app_name: str):
    app_config = apps.get_app_config(app_name)
    urls_py = open(os.path.join(app_config.path, "urls.py"), "r").read()
    return extract_paths(urls_py)


def get_view_type(view: str):
    if view.endswith(ViewType.CREATE.value):
        return ViewType.CREATE

    if view.endswith(ViewType.UPDATE.value):
        return ViewType.UPDATE

    if view.endswith(ViewType.DETAIL.value):
        return ViewType.DETAIL

    if view.endswith(ViewType.LIST.value):
        return ViewType.LIST
    else:
        return None



import os

from django.core.management.base import BaseCommand
from rich.console import Console

from hogwarts.magic_urls.gen_urls import UrlGenerator, urlpatterns_is_empty, UrlMerger
from .base import get_app_config, get_views_module


console = Console()


class Command(BaseCommand):
    help = "urlpatterns generation from views.py"

    def add_arguments(self, parser):
        parser.add_argument("app_name", type=str)

        parser.add_argument(
            "--force-app-name", "-fan",
            action="store_true",
            help="use given app name rather than app_name in urls.py"
        )

        parser.add_argument(
            "--override", "-o",
            action="store_true",
            help="fully overrides urls.py"
        )

        parser.add_argument(
            "--single-import", "-s",
            action="store_true",
            help='import view like "from . import view" instead of importing individual view'
        )

    def handle(self, *args, **options):
        app_name: str = options["app_name"]
        force_new_app_name: bool = options["force_app_name"]
        override: bool = options["override"]
        single_import: bool = options["single_import"]

        views_module = get_views_module(app_name)
        app_config = get_app_config(app_name)
        urls_path = os.path.join(app_config.path, "urls.py")

        url_generator = UrlGenerator(views_module, urls_path, app_name, force_new_app_name, single_import)
        url_merger = UrlMerger(views_module, urls_path, app_name, force_new_app_name, single_import)

        code = ""
        if os.path.exists(urls_path):
            code = open(urls_path, "r").read()

        if not urlpatterns_is_empty(code) and not override:

            url_merger.merge_urls_py()
            console.print("existing paths detected 📜", style="bright_black")
            console.print("adding new paths...", style="bright_black")
            console.print("new paths merged to urlpatterns ✅", style="green")

        else:
            url_generator.gen_urls_py()
            console.print("adding new paths...", style="bright_black")
            console.print("urlpatterns have been generated ✅", style="green")




import os
from typing import Optional

from rich.console import Console
from django.core.management.base import BaseCommand, CommandError

from .base import get_app_config
from hogwarts.magic_views import ViewGenerator
from ...utils import parse_class_names


console = Console()

class Command(BaseCommand):
    help = "Code generation command"

    def add_arguments(self, parser):
        parser.add_argument("app", type=str)
        parser.add_argument("model", type=str)
        parser.add_argument(
            "--smart-mode", "-s",
            action="store_true",
            help="sets login required to create and update actions"
        )

        parser.add_argument(
            "--model-is-namespace", "-mn",
            action="store_true",
            help="sets app name as namespace and action as endpoint for edit views."
                 "see https://docs.djangoproject.com/en/4.2/topics/http/urls/#reversing-namespaced-urls"
                 "or try for yourself"
        )

        parser.add_argument(
            "--file", "-f",
            help="What views file it should read and write"
        )

    def handle(self, *args, **options):
        app_name: str = options["app"]
        model_name: str = options["model"]
        smart_mode: bool = options["smart_mode"]
        model_is_namespace: bool = options["model_is_namespace"]
        views_file: Optional[str] = options["file"]

        app_config = get_app_config(app_name)

        model = app_config.models.get(model_name.lower())
        if model is None:
            raise CommandError(f"Provided model '{model_name}' does not exist in app '{app_name}'")

        namespace_model = False
        if model_is_namespace or model_name.lower() in app_name:
            namespace_model = True

        path = os.path.join(app_config.path, views_file or "views.py")
        self.create_path_if_not_exists(path)

        generator = ViewGenerator(model, smart_mode, namespace_model)

        with open(path, "r") as file:
            existing_code = file.read()
            is_empty = len(parse_class_names(existing_code)) == 0
            if not is_empty:
                console.print("existing views detected 📜 (merging new views)", style="bright_black")
                generator = ViewGenerator(model, smart_mode, namespace_model, code=existing_code)

        code = generator.gen()

        with open(path, 'w') as file:
            file.write(code)

        self.stdout.write(
            self.style.SUCCESS(f"Generated CRUD views in {path}")
        )

    def create_path_if_not_exists(self, path):
        directory = os.path.dirname(path)

        if not os.path.exists(directory):
            os.makedirs(directory)
            console.print("")

        if not os.path.exists(path):
            open(path, 'a').close()



from django.core.management.base import BaseCommand, CommandError
from django.core.management import call_command

from .base import get_app_config


class Command(BaseCommand):
    help = "scaffold model (generate views, urls and templates at once)"

    def add_arguments(self, parser):
        parser.add_argument("app_name", type=str)

    def handle(self, *args, **options):
        app_name = options["app_name"]

        app_config = get_app_config(app_name)

        for model in app_config.get_models():
            try:
                call_command("genviews", app_name, model.__name__, "-s", "-mn")
            except CommandError as e:
                self.stderr.write(self.style.ERROR(f"Could not generate views for model {model.__name__}: {e}"))

        try:
            call_command("genurls", app_name, "-s")
        except CommandError as e:
            self.stderr.write(self.style.ERROR(f"Could not generate urls: {e}"))

        try:
            call_command("gentemplates", app_name)
        except CommandError as e:
            self.stderr.write(self.style.ERROR(f"Could not generate templates: {e}"))



import os

from django.apps import apps
from django import forms
from django.views.generic import FormView, RedirectView

from hogwarts.magic_templates.gen_templates import Endpoint, get_endpoint, ViewType
from hogwarts.magic_urls.gen_urls import get_app_name
from hogwarts.magic_urls.utils import extract_paths
from hogwarts.views import ExampleCreateView
from hogwarts.models import Article

base_path = apps.get_app_config("hogwarts").path
urls_code = open(os.path.join(base_path, "urls.py"), "r").read()

paths = extract_paths(urls_code)
app_name = get_app_name(urls_code)


def test_it_gets_endpoint():
    endpoint = get_endpoint(ExampleCreateView, paths, app_name)

    expected = Endpoint(
        view=ExampleCreateView,
        template_name=ExampleCreateView.template_name,
        path_name="example:create",
        view_type=ViewType.CREATE,
        model=ExampleCreateView.model
    )

    assert endpoint == expected


def test_it_tries_to_get_model_from_form_class():
    class ArticleForm(forms.ModelForm):
        class Meta:
            model = Article
            fields = ["title", "description"]

    class ArticleFormView(FormView):
        form_class = ArticleForm
        template_name = "form.html"
        success_url = "/"

    endpoint = get_endpoint(ArticleFormView, paths, app_name)

    assert endpoint.model == Article


def test_it_sets_null_if_no_model():
    class MyRedirectView(RedirectView):
        def get_redirect_url(self, *args, **kwargs):
            return "/"

    endpoint = get_endpoint(MyRedirectView, paths, app_name)

    assert endpoint.model is None



import ast


def to_plural(string: str):
    if string.endswith("s"):
        return string
    elif string.endswith("y"):
        return string[:-1] + "ies"
    else:
        return string + "s"


def get_imports_list(path: str) -> list[str]:
    with open(path, "r") as file:
        tree = ast.parse(file.read())

    imports = []

    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                imports.append(alias.name)
        elif isinstance(node, ast.ImportFrom):
            for alias in node.names:
                imports.append(alias.name)

    return imports


def get_classes_list(path: str) -> list[str]:
    with open(path, "r") as file:
        tree = ast.parse(file.read())

    classes = []

    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef):
            classes.append(node.name)

    return classes


def code_strip(code: str):
    if code[0] == '\n':
        code = code[1:]

    spaces = 0
    code = code.split("\n")

    for i in code[0]:
        if i == " ":
            spaces += 1
        else:
            break

    for i in range(len(code)):
        code[i] = code[i][spaces:].rstrip()

    return "\n".join(code)


def remove_empty_lines(text: str):
    lines = text.splitlines()
    non_empty_lines = [line for line in lines if line.strip()]  # DON'T TOUCH THIS LINE
    return '\n'.join(non_empty_lines)


def parse_class_names(code):
    class_names = []

    tree = ast.parse(code)

    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef):
            class_names.append(node.name)

    return class_names


def remove_imports(code):
    lines = code.split('\n')
    new_lines = []
    inside_import_block = False

    for line in lines:
        if line.strip().startswith("import ") or line.strip().startswith("from "):
            inside_import_block = True
        elif inside_import_block and not line.strip():
            inside_import_block = False
            continue

        if not inside_import_block:
            new_lines.append(line)

    new_code = '\n'.join(new_lines)
    return new_code



from hogwarts.magic_urls.base import get_path_url


def test_get_path_url():
    assert get_path_url("list", "none", False) == ""
    assert get_path_url("detail", "none", True) == "<int:pk>/"
    assert get_path_url("like", "none", True) == "<int:pk>/like/"
    assert get_path_url("post_detail", "Post", True) == "posts/<int:pk>/"
    assert get_path_url("post_create", "Post", False) == "posts/create/"
    assert get_path_url("post_remove_likes", "Post", False) == "posts/remove-likes/"
    assert get_path_url("post_like", "Post", True) == "posts/<int:pk>/like/"
    assert get_path_url("post_list", "Post", False) == "posts/"




from django.core.management.base import BaseCommand


from ...magic_templates.gen_templates import gen_templates

class Command(BaseCommand):
    help = "Code generation command"

    def add_arguments(self, parser):
        parser.add_argument("app", type=str)

    def handle(self, *args, **options):
        app_name: str = options["app"]
        gen_templates(app_name)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hogwarts',
 'hogwarts.magic_templates',
 'hogwarts.magic_urls',
 'hogwarts.magic_views',
 'hogwarts.management',
 'hogwarts.management.commands',
 'hogwarts.migrations',
 'hogwarts.tests',
 'hogwarts.tests.template_tests',
 'hogwarts.tests.url_tests',
 'hogwarts.tests.view_tests']

package_data = \
{'': ['*'], 'hogwarts': ['scaffold/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0', 'rich>=13.5.2,<14.0.0']

setup_kwargs = {
    'name': 'django-hogwarts',
    'version': '0.8.0',
    'description': 'Django utilities for codegen and DX improvement',
    'long_description': '<h1 align="center">Django hogwarts 🧙\u200d♂️</h1>\n<h4 align="center">Management commands to generate views, urls and templates</h4>\n\nUse CLI commands to generate:\n- basic create, update, list, detail views\n- urlpatterns from views with REST like path urls\n- form, table, detail templates (Bootstrap and django-crispy-forms by default)\n\n**all commands will respect (will not change) existing code**\n\n---\n\n## Installation\n```shell\n# pip\npip install django-hogwarts\n\n# poetry\npoetry add django-hogwarts\n```\n\nadd `hogwarts` to your `INSTALLED_APPS`:\n``` python\nINSTALLED_APPS = [\n    ...\n    "hogwarts"\n]\n```\n\n## Usage\n> Check [this](./docs/conventions.md) to know what urls will be generated\n### Generate urls.py\nGenerates paths for views from views.py\n```\npython manage.py genurls <your-app-name>\n```\n\nArguments:\n- `--force-app-name`, `fan` override app_name variable in urls.py \n- `--override`, `-o` fully overrides existing code in urls.py (previous code will be deleted)\n- `--single-import`, `-s` instead of importing individual view, imports just module`from . import views`\n\n### Generate views.py\nGenerates create, update, detail, list views for model.\nCheckout the [demo](./docs/gen_views_example.md)\n```\npython manage.py genviews <your-app-name> <model-name>\n```\nArguments\n- `--smart-mode`, `-s` adds login required, sets user for CreateView and checks if client is owner of object in UpdateView\n- `--model-is-namespace`, `-mn` adds success_url with name model as [namespace](https://docs.djangoproject.com/en/4.2/topics/http/urls/#url-namespaces)\n- `--file`, `-f` specify view file (example: "views/posts_view.py" or "new_views.py") in your app\n\n### Generate templates\nGenerates templates from `template_name`s from views from given app\n\n**[django-crispy-forms](https://github.com/django-crispy-forms/django-crispy-forms) and\n[crispy-bootstrap5](https://github.com/django-crispy-forms/crispy-bootstrap5) packages are required**\n\n``` \npython manage.py gentemplates <your-app-name>\n```\n\nWant to create own scaffolding templates? \n1. create folder, copy and customize from [this templates](https://github.com/adiletto64/django-hogwarts/tree/master/hogwarts/scaffold)\n2. add that folder to setting `HOGWARTS_SCAFFOLD_FOLDER = "<your-folder>"`\n\n### Scaffolding\n\nGenerates views, urls and templates for given app (every model in app)\n\n``` \npython manage.py scaffold <your-app-name>\n```\n\n\n## Roadmap\n- tests generator\n- maybe rest-framework support (let me know in issues)\n\n\n',
    'author': 'adiletto64',
    'author_email': 'adiletdj19@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/adiletto64/django-hogwarts',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.views.generic import CreateView, DetailView, ListView, UpdateView

from .models import Example


class ExampleListView(ListView):
    model = Example
    template_name = "list.html"
    context_object_name = "examples"


class ExampleCreateView(CreateView):
    model = Example
    fields = ["message"]
    template_name = "create.html"
    success_url = reverse_lazy("example:list")


class ExampleDetailView(DetailView):
    model = Example
    template_name = "detail.html"


class AddEqualSignView(UpdateView):
    model = Example
    fields = ["message"]

    def post(self, request, *args, **kwargs):
        obj: Example = self.get_object()
        obj.message = obj.message + "="
        obj.save()

        return redirect("example:list")



from typing import Optional

from hogwarts.magic_urls.base import Path, get_path_url


def custom_path(
        path_name: str,
        path_url: Optional[str] = None,
        detail: bool = False
):
    def wrapper(obj):
        obj.auto_url_path = Path(path_name, detail, path_url)
        return obj

    return wrapper


class PathDecorator:
    def __init__(self, view):
        self.view = view

    def exists(self):
        return hasattr(self.view, "auto_url_path")

    def get_path_name(self):
        path: Path = getattr(self.view, "auto_url_path")
        return path.path_name

    def get_path_url(self):
        path: Path = getattr(self.view, "auto_url_path")
        path_name = self.get_path_name()
        return path.path_urls or get_path_url(path_name, detail=path.detail)
